/* vim: set sw=8 ts=8 si : */
/*********************************************
* Author: Guido Socher
* Copyright: GPL V2
**********************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include "analog.h"

// if you change MX_CHANNELS then change as well the inittialisation of chpos below
#define MX_CHANNELS 2
// adc measurement results (12bit ADC):
static volatile uint16_t analog_result[MX_CHANNELS];

// 1.1V Vref with capacitor on AREF pin:
#if defined(__AVR_ATmega644__)||defined(__AVR_ATmega644P__)
#define ADMUX_REFSBITS (1<<REFS1)
#else
// atmega168, atmega328, 1.1V:
#define ADMUX_REFSBITS (1<<REFS1)|(1<<REFS0)
#endif
// Pre-scaling is needed to stay in the recommended range of 50-200kHz 
// Clock freq for the adc.
//pre-scale by 128:
//#define ADCSRA_PSBITS (1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0)
// pre-scale by 64 (at 12MHz this will be 187kHz ADC clock):
#define ADCSRA_PSBITS (1<<ADPS2)|(1<<ADPS1)

#ifdef USE_INTERRUPT
// Do not use interrupt driven ADC measurements with
// systems where the timing is critical (e.g DS18S20 sensors).
// The ISR(ADC_vect) interrupt function will mess up you timing
// too much.
void adc_isr_on(void)
{
        /* enable analog to digital conversion in free run mode
        *  without noise canceler function. 
        * ADEN: Analog Digital Converter Enable
        * ADIE: ADC Interrupt Enable
        * ADIF: ADC Interrupt Flag
        * ADCSR: ADC Control and Status Register
        * ADPS2..ADPS0: ADC Prescaler Select Bits
	* REFS: Reference Selection Bits (page 203)
        */
        ADCSRA=(1<<ADEN)|(1<<ADATE)|(1<<ADIE)|(1<<ADIF)|ADCSRA_PSBITS;
        ADCSRB=0;
        //  start conversion
        ADCSRA|=(1<<ADSC);
}

uint16_t getanalogresult(uint8_t channel)
{
	if (channel>=MX_CHANNELS) return 0;
        return(analog_result[channel]);
}

/* the following function will be called when analog conversion is done.
 * It will be called every 13th cycle of the converson clock. At 12.5Mhz
 * and a ADPS factor of 64 this is: ((12500 kHz/ 64) / 13)= 15KHz intervalls
 *
 * We do 16 fold oversampling for 12bit as explained in atmel doc AVR121.
 *
 * This function executes in an atomic way and analog_result[] can
 * not be read in the middle of the execution of this function.
 */
ISR(ADC_vect) {
	static uint8_t channel=0; 
	static uint8_t chpos[MX_CHANNELS]={0,0};
	uint8_t pos;
	// raw 10bit values:
        static uint16_t raw_analog_result[MX_CHANNELS][16];  
	uint8_t i;
	uint8_t adlow;
	// toggel the channel between 0 an 1. This will however
	// not effect the next conversion as that one is already
	// ongoing.
	channel^=1;
	// 
	ADMUX=ADMUX_REFSBITS |channel;
	pos=chpos[channel];
	adlow=ADCL; // read low first !! 
	raw_analog_result[channel][pos]=(ADCH<<8)|adlow;
	// we do 16 bit oversampling to get 12bit ADC resolution
	//
	i=0;
	analog_result[channel]=0;
	while(i<16){
		analog_result[channel]+=raw_analog_result[channel][i];
		i++;
	}
	analog_result[channel]=analog_result[channel]>>2; // 12bit
	pos=(pos+1)%16; // rotate over 16 
	chpos[channel]=pos;
	// end of interrupt handler
}

#else // USE_INTERRUPT

// return analog value of a given channel. Works without interrupt 
// Resolution = 10bit
uint16_t convertanalog(unsigned char channel) 
{
	unsigned char adlow;
	uint16_t result;
        //  enable analog to digital conversion in single run mode
        //  without noise canceler function. See datasheet of atmega88 page 250
        // ADEN: Analog Digital Converter Enable
        // ADIE: ADC Interrupt Enable  (0=no interrupt)
        // ADIF: ADC Interrupt Flag
        // ADCSR: ADC Control and Status Register
        // ADPS2..ADPS0: ADC Prescaler Select Bits
	// REFS: Reference Selection Bits (page 203)
        // 
	// write only the lower 3 bit for channel selection
	ADMUX=ADMUX_REFSBITS |(channel & 0x7);
	// 
	// switch off digital input line:
	if (channel <6){// 6 and 7 have no digital input
		DIDR0=(1<<channel)& 0x3f;
	}
        ADCSRA=(1<<ADEN)|ADCSRA_PSBITS;
	//  start conversion 
	ADCSRA|= (1<<ADSC);
	while(bit_is_set(ADCSRA,ADSC)); // wait for result 
	adlow=ADCL; // read low first
	// use only 10 bit resolution:
	result=(ADCH<<8)|adlow;
	return(result);
}

#endif // USE_INTERRUPT
